@@ -114,6 +114,9 @@ class Events:
                 args += ","
         return self.__reply_status_check(self.device.con.get(path="sendSlideEvent", args="sliders=" + args))
 
+    def power(self, delay=0):
+        return self.__reply_status_check(self.device.con.get(path="sendPowerKeyEvent", args="delay=" + str(delay)))
+
     def back(self, delay=0):
         return self.__reply_status_check(self.device.con.get(path="sendBackKeyEvent", args="delay=" + str(delay)))
 
@@ -216,6 +219,15 @@ class Events:
     def system_time(self):
         return int(str(self.device.con.get(path="getDatetime").read(), 'utf-8'))
 
+    def set_system_time(self, secs: int):
+        return self.__reply_status_check(self.device.con.get(path="setDatetime", args="datetime=" + str(secs)))
+
+    def get_system_auto_time(self):
+        return bool(int(str(self.device.con.get(path="getAutoDatetime").read(), 'utf-8')))
+
+    def set_system_auto_time(self, state: bool):
+        return self.__reply_status_check(self.device.con.get(path="setAutoDatetime", args="state=" + str(int(state))))
+
     def latest_toast(self):
         return str(self.device.con.get(path="getLatestToast").read(), 'utf-8')
 
