@@ -13,8 +13,10 @@
 #
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
+import base64
 import time
-
+from ctypes import *
+import threading
 import sympy as sp
 import math
 from lxml import etree
@@ -23,6 +25,29 @@ import xml.dom.minidom
 import numpy as np
 import random
 from time import sleep
+import io
+
+
+html_string_1 = "<!DOCTYPE html><html lang=\"en\"><head><meta charset=\"UTF-8\"><title>render</title></head><body style=\"background:#000;clip:rect(auto,auto,auto,auto);\"><div id=\"content\" style=\"width:"
+html_string_2 = "px;height:"
+html_string_3 = "px;clip:rect(auto,auto,auto,auto);\">"
+html_string_4 = "</div></body></html>"
+
+
+class WorkerThread(threading.Thread):
+
+    result = None
+
+    def __init__(self, func, args=()):
+        super(WorkerThread, self).__init__()
+        self.func = func
+        self.args = args
+
+    def run(self) -> None:
+        self.result = self.func(*self.args)
+
+    def get_result(self):
+        return self.result
 
 
 class Item:
@@ -202,28 +227,39 @@ class Item:
     def exist(self, timeout=None):
         if timeout is None:
             timeout = self.device.default_timeout
-        sleep(3)
-        tree = xml.dom.minidom.parseString(self.device.xmlStr)
-        if tree.documentElement.getAttribute("sopId") == self.sopid:
-            die_time = int(time.time()) + timeout
-            while int(time.time()) < die_time:
-                self.rect = self.__exist()
-                if len(self.rect) > 0:
-                    return True
-                sleep(0.5)
+        die_time = int(time.time()) + timeout
+        while True:
+            self.rect = self.__exist()
+            if len(self.rect) > 0:
+                return True
+            sleep(0.5)
+            if int(time.time()) > die_time:
+                break
         return False
 
     def __exist(self):
         self.__refresh_node()
         if self.node is not None and self.__visible and self.__opacity > 0 and self.__scale != 0 and self.__width > 0 and self.__height > 0:
             image = []
-            tree = xml.dom.minidom.parseString(self.device.xmlStr)
-            for node in tree.documentElement.childNodes:
-                for n in node.childNodes:
-                    if len(image) == 0:
-                        image = self.__xml_tree_traversed(n.childNodes, 1)
-                    else:
-                        image = self.__img_cover(image, self.__xml_tree_traversed(n.childNodes, 1))
+            if self.device.control_host_type == 0:
+                tree = xml.dom.minidom.parseString(self.device.xmlStr)
+                for node in tree.documentElement.childNodes:
+                    for n in node.childNodes:
+                        if len(image) == 0:
+                            image = self.__xml_tree_traversed(n.childNodes, 1)
+                        else:
+                            image = self.__img_cover(image, self.__xml_tree_traversed(n.childNodes, 1))
+            else:
+                html_string = self.device.libsr.go(c_char_p(bytes(self.device.xmlStr, "utf-8")),
+                                                   c_char_p(bytes(self.__tempId, "utf-8")),
+                                                   c_int(self.__display_width),
+                                                   c_int(self.__display_height))
+                self.device.webdriver.get("data:text/html;charset=utf-8," +
+                                          html_string_1 + str(self.__display_width) +
+                                          html_string_2 + str(self.__display_height) +
+                                          html_string_3 + string_at(html_string, -1).decode('utf-8') +
+                                          html_string_4)
+                image = cv2.imdecode(np.frombuffer(base64.b64decode(self.device.webdriver.get_screenshot_as_base64()), np.uint8), cv2.COLOR_RGB2BGR)
             # cv2.imshow(str(self.device.system_time()), image)
             # cv2.waitKey(0)
             b, g, r = cv2.split(image)
@@ -232,61 +268,82 @@ class Item:
         return []
 
     def __xml_tree_traversed(self, nodes, fs):
-        images = {}
+        all_images = {}
+        th_index = 0
+        th_list = {}
         for node in nodes:
-            clip = int(node.getAttribute("clip"))
-            scale = abs(float(node.getAttribute("scale")) * fs)
-            height = round(float(node.getAttribute("height")) * scale)
-            width = round(float(node.getAttribute("width")) * scale)
-            rotation = 0 - int(node.getAttribute("rotation"))
-            has_contents = bool(int(node.getAttribute("hasContents")))
-            cx = round(float(node.getAttribute("centerXToGlobal")))
-            cy = round(float(node.getAttribute("centerYToGlobal")))
-            if height > 10000 or width > 10000:
-                x = float(node.getAttribute("x"))
-                y = float(node.getAttribute("y"))
-                cx, cy, width, height = self.__clip_item(float(cx), float(cy), float(width), float(height), x, y, 0 - rotation)
-            if float(node.getAttribute("opacity")) != 0 \
-                    and int(node.getAttribute("visible")) == 1 \
-                    and node.nodeName != "QQuickShaderEffectSource":
-                index_z = int(node.getAttribute("z"))
-                if height > 0 and width > 0:
-                    image = np.zeros((height, width), dtype=np.uint8)
+            th = WorkerThread(func=self.__draw_item, args=(node, fs, ))
+            th_list[th_index] = th
+            th.setDaemon(True)
+            th_index += 1
+        for i in range(th_index):
+            th_list[i].start()
+        for i in range(th_index):
+            th_list[i].join()
+            images = th_list[i].get_result()
+            for image_z in images:
+                if image_z in all_images.keys():
+                    all_images[image_z] = self.__img_cover(all_images[image_z], images[image_z])
                 else:
-                    image = np.zeros((self.__display_height, self.__display_width), dtype=np.uint8)
-                image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
-                if has_contents or (node.nodeName == "QQuickLoader" and node.getAttribute("z") == "10000" and
-                                    node.getAttribute("tempID") == "temp7"):
-                    if node.getAttribute("tempID") == self.__tempId:
-                        cv2.rectangle(image, (0, 0), (width, height), (0, 0, 255), -1)
-                    else:
-                        # cv2.rectangle(image, (0, 0), (width, height), (255, 0, 0), -1)
-                        cv2.rectangle(image, (0, 0), (width, height),
-                                     (random.randint(0, 255), 0, random.randint(0, 255)), -1)
-                else:
-                    cv2.rectangle(image, (0, 0), (width, height), (0, 255, 0), -1)
-                hypotenuse = int(math.sqrt(height ** 2 + width ** 2))
-                M1 = np.float32([[1, 0, (hypotenuse - width) / 2], [0, 1, (hypotenuse - height) / 2]])
-                image = cv2.warpAffine(image, M1, (hypotenuse, hypotenuse))
-                M2 = cv2.getRotationMatrix2D((hypotenuse / 2, hypotenuse / 2), rotation, 1)
-                image = cv2.warpAffine(image, M2, (hypotenuse, hypotenuse))
-                M3 = np.float32([[1, 0, cx - (hypotenuse / 2)], [0, 1, cy - (hypotenuse / 2)]])
-                image = cv2.warpAffine(image, M3, (self.__display_width, self.__display_height))
-                if node.childNodes.length > 0:
-                    c_images = self.__xml_tree_traversed(node.childNodes, scale)
-                    image = self.__img_cover(image, c_images, clip)
-                if index_z in images.keys():
-                    images[index_z] = self.__img_cover(images[index_z], image)
-                else:
-                    images[index_z] = image
-        im_list = list(images.keys())
+                    all_images[image_z] = images[image_z]
+        im_list = list(all_images.keys())
         im_list.sort()
         r_image = np.zeros((self.__display_height, self.__display_width), dtype=np.uint8)
         r_image = cv2.cvtColor(r_image, cv2.COLOR_GRAY2BGR)
         for index in im_list:
-            r_image = self.__img_cover(r_image, images[index])
+            r_image = self.__img_cover(r_image, all_images[index])
         return r_image
 
+    def __draw_item(self, node, fs):
+        images = {}
+        clip = int(node.getAttribute("clip"))
+        scale = abs(float(node.getAttribute("scale")) * fs)
+        height = round(float(node.getAttribute("height")) * scale)
+        width = round(float(node.getAttribute("width")) * scale)
+        rotation = 0 - int(node.getAttribute("rotation"))
+        has_contents = bool(int(node.getAttribute("hasContents")))
+        cx = round(float(node.getAttribute("centerXToGlobal")))
+        cy = round(float(node.getAttribute("centerYToGlobal")))
+        if height > 10000 or width > 10000:
+            x = float(node.getAttribute("x"))
+            y = float(node.getAttribute("y"))
+            cx, cy, width, height = self.__clip_item(float(cx), float(cy), float(width), float(height), x, y,
+                                                     0 - rotation)
+        if float(node.getAttribute("opacity")) != 0 \
+                and int(node.getAttribute("visible")) == 1 \
+                and node.nodeName != "QQuickShaderEffectSource":
+            index_z = int(node.getAttribute("z"))
+            if height > 0 and width > 0:
+                image = np.zeros((height, width), dtype=np.uint8)
+            else:
+                image = np.zeros((self.__display_height, self.__display_width), dtype=np.uint8)
+            image = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
+            if has_contents or (node.nodeName == "QQuickLoader" and node.getAttribute("z") == "10000" and
+                                node.getAttribute("tempID") == "temp7"):
+                if node.getAttribute("tempID") == self.__tempId:
+                    cv2.rectangle(image, (0, 0), (width, height), (0, 0, 255), -1)
+                else:
+                    cv2.rectangle(image, (0, 0), (width, height), (255, 0, 0), -1)
+                    # cv2.rectangle(image, (0, 0), (width, height),
+                    #              (random.randint(0, 255), 0, random.randint(0, 255)), -1)
+            else:
+                cv2.rectangle(image, (0, 0), (width, height), (0, 255, 0), -1)
+            hypotenuse = int(math.sqrt(height ** 2 + width ** 2))
+            M1 = np.float32([[1, 0, (hypotenuse - width) / 2], [0, 1, (hypotenuse - height) / 2]])
+            image = cv2.warpAffine(image, M1, (hypotenuse, hypotenuse))
+            M2 = cv2.getRotationMatrix2D((hypotenuse / 2, hypotenuse / 2), rotation, 1)
+            image = cv2.warpAffine(image, M2, (hypotenuse, hypotenuse))
+            M3 = np.float32([[1, 0, cx - (hypotenuse / 2)], [0, 1, cy - (hypotenuse / 2)]])
+            image = cv2.warpAffine(image, M3, (self.__display_width, self.__display_height))
+            if node.childNodes.length > 0:
+                c_images = self.__xml_tree_traversed(node.childNodes, scale)
+                image = self.__img_cover(image, c_images, clip)
+            if index_z in images.keys():
+                images[index_z] = self.__img_cover(images[index_z], image)
+            else:
+                images[index_z] = image
+        return images
+
     def __clip_item(self, cx, cy, width, height, x, y, rotation):
         if width > 10000:
             r_width = 10000
